#python
from numpy import *
from scipy import sparse
from collections import defaultdict

from competition import Classifier, UntrainedError, scoreClassifier



class ExampleClassifier(Classifier):
    """
    A very naive "decision stump" classifier that tries to predict from the 
    presence of a single word.
    """
    def __init__(self, sname="Monty", cname="decision stump", about="version 1.0", debug=True):
        # first, send identifier into to the base class
        super(ExampleClassifier, self).__init__(sname, cname, about)
        
        self.debug = debug      # print out extra debug info
        self.feature = None     # the feature we use to assign the label
        self.truelabel = None   # the label we assign when feature > 0
        self.falselabel = None  # the label we assign when feature = 0
        
        
    def train(self, train):
        """
        Walk through the dimensions of the data and find the one for which 
        having val > 0 best predicts the label.
        
        'train' is the data in the form of a sequence of (label, data) pairs
        """
        X, feats = self.text2array([dat for _,dat in train])
        Y = array([y for y,_ in train])
        
        bestscore = -1
    
        NX, NA = X.shape
        assert NA==len(feats)
        
        for a in xrange(NA):
            # we compute score by adding the labels of all the nonzero entries
            # in the column.  if score is high or low, there is a difference 
            # between the positive and negative labels values.
            score = sum(Y[X[:,a].nonzero()[0]])
            if self.debug:
                print '[train] feature "%s", score = %d' % (feats[a], score)

            if abs(score) > bestscore:
                bestscore = abs(score)
                self.feature = feats[a]
                if score > 0:
                    self.truelabel = 1
                    self.falselabel = -1
                else:
                    self.truelabel = -1
                    self.falselabel = 1
                if self.debug:
                    print '[train] best feature = %s [abs(score) = %d]' % (self.feature, bestscore)
        # print '***** best feature =', self.feature
        
                    
    def test(self, teststrings):
        """
        Return the label hypotheses for test strings.  Note that because our
        decision stump classifier only looks for the presence of a feature, we
        can test directly on the string, without converting to an array or 
        matrix.
        """
        if self.feature is None:
            # we haven't been trained!
            raise UntrainedError
        
        H = []
        for tstr in teststrings:
            s = ''.join([c if c.isalpha() else ' ' for c in tstr.lower()])
            tokens = s.split()            
            if self.feature in tokens:
                H.append(self.truelabel)
            else:
                H.append(self.falselabel)
        return array(H)


    def text2array(self, strings):
        """ 
        given a sequence of strings create a sparse array for a bag-of-words 
        model 
        """
        # first, we will do some simple processing: replace anything that isn't
        # a letter with a space, convert to lowercase, and tokenize
        tokenlists = []
        fcount = defaultdict(int)  # features are the various tokens (mostly 
                                   # words). defaultdict is a handy way of
                                   # counting occurrences (among other things)
        for s in strings:
            s = ''.join([c if c.isalpha() else ' ' for c in s.lower()])
            tokens = s.split()
            for tok in tokens:
                fcount[tok] += 1
            tokenlists.append(tokens)
            if self.debug:
                print "[processing] made string into sequence of %d tokens" % len(tokens)
        
        # make a map from feature -> array_index, but only use features that
        # occured in at least 5% of the data
        thresh = len(strings) / 20
        features = [f for f,c in fcount.iteritems() if c > thresh]
        feat2ind = dict((f,i) for i,f in enumerate(features))
        
        # the data array is 2D and contains the count of features (words) for
        # each document.  X[i,j] is the number of times feature j appears in
        # document i.  clearly, this is sparse, so we'll use scipy's sparse
        # matrix data structure to save memory
        X = sparse.lil_matrix((len(strings), len(features)))
        for i, toks in enumerate(tokenlists):
            for t in toks:
                if t in features:
                    X[i,feat2ind[t]] += 1
            if self.debug:
                print '[processing] populated row %d' % i
        
        return X, features
            

#############################################################################
#  The file you submit should have the following code at the bottom,
#  unmodified, except to use your classifier.
#############################################################################
def testClassifier():
    
    cfr = ExampleClassifier(debug=False)
    scoreClassifier(cfr)
    
    
    
if __name__ == "__main__":
    # run the script from the cammand line:
    #
    # > python exampleclassifier.py
    #
    testClassifier()